import logging
import re
import sys

from collections import OrderedDict as OrdDict
from collections import namedtuple
from operator import itemgetter, attrgetter
from avclass import DEFAULT_TAX_PATH, DEFAULT_TAG_PATH, DEFAULT_EXP_PATH

# Set logging
log = logging.getLogger(__name__)

# Prefix to identify platform tags
platform_prefix = "FILE:os:"

# Default category for tags in taxonomy with no category
uncategorized_cat  = "UNC"

SampleInfo = namedtuple('SampleInfo', 
                        ['md5', 'sha1', 'sha256', 'labels', 'vt_tags'])

# AVs to use in suffix removal
suffix_removal_av_set = {
    'Norman',
    'Avast',
    'Avira',
    'Kaspersky',
    'ESET-NOD32',
    'Fortinet',
    'Jiangmin',
    'Comodo',
    'GData',
    'Sophos',
    'TrendMicro-HouseCall',
    'TrendMicro',
    'NANO-Antivirus',
    'Microsoft'
}

class Tag:
    """A Tag in the taxonomy"""
    def __init__(self, s):
        word_list = s.strip().split(":")
        if len(word_list) > 1:
            self._name = word_list[-1].lower()
            self._cat = word_list[0].upper()
            self._prefix_l = [x.lower() for x in word_list[1:-1]]
            path = self._cat
            for x in self._prefix_l:
                path = path + ':' + x
            self._path = path + ':' + self._name
        else:
            self._name = word_list[0].lower()
            self._cat = uncategorized_cat
            self._prefix_l = []
            self._path = self._name

    def __hash__(self):
        """Return hash"""
        return hash((self._path))

    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    @property
    def name(self):
        """Return tag name"""
        return self._name

    @property
    def cat(self):
        """Return tag category"""
        return self._cat

    @property
    def path(self):
        """Return tag path"""
        return self._path

    @property
    def prefix_l(self):
        """Return tag prefix list"""
        return self._prefix_l


class Taxonomy:
    """A taxonomy of tags and generic tokens read from file"""
    def __init__(self, filepath):
        """Initialize and populate _tag_map from input file"""
        self._tags = set() # Map tag.name | tag.path -> Tag
        self._tag_map = {}
        if filepath:
            self.read_taxonomy(filepath)

    def __len__(self):
        """Taxonomy length is the number of tags it contains"""
        return len(self._tags)

    def __iter__(self):
        """Iterator over the alphabetically sorted tags in the taxonomy"""
        return (t for t in sorted(self._tags))

    def is_generic(self, t):
        """Whether the input tag is generic"""
        tag = self._tag_map.get(t, None)
        if tag:
            return tag.cat == "GEN"
        else:
            return False

    def is_tag(self, t):
        """Whether the input tag exists in the taxonomy"""
        return t in self._tag_map

    def add_tag(self, s, override=False):
        """Add the input tag to the taxonomy 

        If tag already exists with different path, 
        it is only replaced if override is True
        """
        tag = Tag(s)
        t = self._tag_map.get(tag.name, None)
        if t and (t.path != tag.path):
            if (not override):
                return
            else:
                log.warning("[Taxonomy] Replacing %s with %s\n" % (
                                  t.path, tag.path))
                del self._tag_map[t.path]
        log.debug("[Taxonomy] Adding tag %s" % s)
        self._tags.add(tag)
        self._tag_map[tag.name] = tag
        self._tag_map[tag.path] = tag
        return

    def remove_tag(self, t):
        """Remove tag from taxonomy. Returns whether tag was removed"""
        tag = self._tag_map.get(t, None)
        if tag:
            log.debug("[Taxonomy] Removing tag: %s" % tag.path)
            del self._tag_map[tag.name]
            del self._tag_map[tag.path]
            self._tags.remove(tag)
        return tag is not None

    def get_category(self, t):
        """Return category of input tag, UNK if not a tag"""
        tag = self._tag_map.get(t, None)
        if tag:
            return tag.cat
        else:
            return "UNK"

    def get_path(self, t):
        """Return full path for given tag, or empty string if not a tag"""
        tag = self._tag_map.get(t, None)
        if tag:
            return tag.path
        else:
            return ("UNK:" + t)

    def get_prefix_l(self, t):
        """Return prefix list for given tag, or empty string if not a tag"""
        tag = self._tag_map.get(t, None)
        if tag:
            return tag.prefix_l
        else:
            return []

    def get_prefix(self, t):
        """Return prefix string for given tag, or empty string if not a tag"""
        tag = self._tag_map.get(t, None)
        if tag:
            return tag.prefix_l
        else:
            return t.path[0:t.path.rfind(':')]

    def get_depth(self, t):
        """Return depth of tag in taxonomy. 

        Returns zero if tag not in taxonomy. 
        A normal tag CAT:name has depth two
        """
        tag = self._tag_map.get(t, None)
        if tag:
            return len(tag.prefix_l) + 2
        else:
            return 0

    def get_info(self, t):
        """Return (path,category) for given tag, or UNK:t if not a tag"""
        tag = self._tag_map.get(t, None)
        if tag:
            return tag.path, tag.cat
        else:
            return "UNK:" + t, "UNK"

    def expand(self, t):
        """Return list of tags in prefix list that are leaves"""
        tag = self._tag_map.get(t, None)
        if tag:
            return [t for t in tag.prefix_l if t in self._tag_map]
        else:
            return []

    def platform_tags(self): 
        """Returns list with platform tags in taxonomy"""
        acc = set()
        for idx,tag in self._tag_map.items():
            if tag.path.startswith(platform_prefix):
                acc.add(tag.name)
        return acc

    def overlaps(self, t1, t2):
        """Returns true if the path of the given tags overlaps"""
        m1 = self.get_prefix_l(t1)
        m2 = self.get_prefix_l(t2)
        return (t1 in m2) or (t2 in m1)

    def remove_overlaps(self, l): 
        """Returns list with overlapping tags removed"""
        if not l:
            return l
        pair_l = sorted([(self.get_depth(t),t) for t in l])
        out_l = [pair_l.pop()[1]]
        while pair_l:
            t = pair_l.pop()[1]
            if (not any(self.overlaps(t, e) for e in out_l)):
                out_l.append(t)
        return out_l

    def read_taxonomy(self, filepath):
        """Read taxonomy from given file"""
        with open(filepath, 'r') as fd:
            for line in fd:
                if line.startswith('#') or line == '\n':
                    continue
                self.add_tag(line.strip())
        return

    def to_file(self, filepath):
        """Output sorted taxonomy to given file"""
        # Open output file
        fd = open(filepath, 'w')
        # Write sorted tags
        tag_l = sorted(self._tag_map.items(), 
                                key=lambda item : item[1].path, 
                                reverse=False)
        idx = 0
        for name,tag in tag_l:
            if (idx % 2) == 0:
                fd.write(tag.path+"\n")
            idx+=1
        # Close output file
        fd.close()

class Rules:
    """A relation from one source to one or more destinations"""
    def __init__(self, filepath):
        """Initialize rule map and read rules from input file"""
        self._src_map = {} # src -> set(dst)
        if filepath:
            self.read_rules(filepath)

    def __len__(self):
        """The number of rules (i.e., source tags)"""
        return len(self._src_map)

    def add_rule(self, src, dst_l, overwrite=False):
        """Add rule.

        If rule exists:
        if overwrite==True, replace destination list
        else append dst_l to current target set
        """
        # Remove src from dst_l if it exists
        dst_l = filter(lambda x: x != src, dst_l)
        # If no destinations, nothing to do
        if (not dst_l):
            return
        log.debug("[Rules] Adding %s -> %s" % (src, dst_l))
        src_tag = Tag(src)
        if overwrite:
            target_l = [Tag(dst).name for dst in dst_l]
            self._src_map[src_tag.name] = set(target_l)
        else:
            curr_dst = self._src_map.get(src_tag.name, set())
            for dst in dst_l:
                dst_tag = Tag(dst)
                curr_dst.add(dst_tag.name)
            self._src_map[src_tag.name] = curr_dst
        return

    def remove_rule(self, src):
        """Remove the rule for the input source"""
        l = self._src_map.get(src, [])
        if l:
            log.debug("[Rules] Removing rule: %s -> %s" % (src, l))
            del self._src_map[src]
            return 1
        else:
            return 0

    def get_dst(self, src):
        """Returns dst list for given src, or empty list if no expansion"""
        return list(self._src_map.get(src, []))

    def read_rules(self, filepath):
        """Read rules from given file"""
        with open(filepath, 'r') as fd:
            for line in fd:
                if line.startswith('#') or line == '\n':
                    continue
                word_list = line.strip().split()
                if len(word_list) > 1:
                    self.add_rule(word_list[0],word_list[1:])
        return

    def to_file(self, filepath, taxonomy=None):
        """Output sorted rules to given file
 
        If taxonomy is provided, it outputs full tag path
        """
        fd = open(filepath, 'w')
        for src,dst_set in sorted(self._src_map.items()):
            dst_l = sorted(dst_set, reverse=False)
            if taxonomy:
                src_path = taxonomy.get_path(src)
                path_l = [taxonomy.get_path(t) for t in dst_l]
                dst_str = '\t'.join(path_l)
                fd.write("%s\t%s\n" % (src_path,dst_str))
            else:
                dst_str = '\t'.join(dst_l)
                fd.write("%s\t%s\n" % (src,dst_str))
        fd.close()

    def expand_src_destinations(self, src):
        """Return a list of all expanded destinations for given source

           Recursively follows any rules for destinations
        """
        dst_set = self._src_map.get(src, set())
        out = set()
        while dst_set:
            dst = dst_set.pop()
            l = self._src_map.get(dst, [])
            if l:
                for e in l:
                    if (e not in out) and (e != dst):
                        dst_set.add(e)
            else:
                out.add(dst)
        return out

    def expand_all_destinations(self):
        """Expand all sources"""
        src_l = self._src_map.keys()
        for src in src_l:
            dst_l = self.expand_src_destinations(src)
            self._src_map[src] = dst_l

class Tagging(Rules):
    """A rule with an unknown source and a destination in the taxonomy"""
    def __init__(self, filepath):
        """Initialize rules from input file"""
        Rules.__init__(self, filepath)

    def validate(self, taxonomy):
        """Check that tags in tagging rules are in given taxonomy"""
        for src,dst_l in self._src_map.items():
            if taxonomy.is_tag(src):
                log.warning("[Tagging] SRC %s in taxonomy" % src)
            for t in dst_l:
                if (not taxonomy.is_tag(t)):
                    log.warning("[Tagging] %s not in taxonomy" % t)

    def build_synonymn_map(self):
        """Build a map from dst to src set from Tagging"""
        dst_map = {}
        for src, dst_set in self._src_map.items():
            for dst in dst_set:
                dst_map.setdefault(dst, set()).add(src)
        return dst_map

class Expansion(Rules):
    """A rule where source different than destination and both in taxonomy"""
    def __init__(self, filepath):
        Rules.__init__(self, filepath)

    def validate(self, taxonomy):
        """Check that tags in expansion rules are in given taxonomy"""
        for src,dst_set in self._src_map.items():
            if (not taxonomy.is_tag(src)):
                log.warning("[Expansion] %s not in taxonomy" % src)
            for dst in dst_set:
                if (not taxonomy.is_tag(dst)):
                    log.warning("[Expansion] %s not in taxonomy" % dst)

class AvLabels:
    """Primary class to process AV labels"""
    def __init__(
        self,
        tag_file=DEFAULT_TAG_PATH,
        exp_file = DEFAULT_EXP_PATH,
        tax_file = DEFAULT_TAX_PATH,
        av_l = None
    ):
        """Initialize using given files and options"""
        # Read taxonomy
        self.taxonomy = Taxonomy(tax_file)
        # Read tag rules
        self.tagging = Tagging(tag_file)
        # Read expansion rules
        self.expansions = Expansion(exp_file)
        # List of AV engines to use
        self.avs = av_l

    @staticmethod
    def is_pup(tag_pairs, taxonomy):
        """Whether the sample is PUP
 
           Checks if highest ranked CLASS tag contains "grayware"
           and is above a predefined threshold
           It returns False unless it can determine it is PUP
        """
        threshold = 0.5
        # If no tags, return false
        if len(tag_pairs) < 1:
            return False
        max_ctr = tag_pairs[0][1]
        for (tag,ctr) in tag_pairs:
            (path, cat) = taxonomy.get_info(tag)
            if (cat == "CLASS"):
                if ("grayware" in path):
                    return (float(ctr) >= float(max_ctr)*threshold)
                else:
                    return False
        return False

    @staticmethod
    def _remove_suffixes(av_name, label):
        """Returns input label without AV specific suffixes""" 
        # Truncate after last '.'
        if av_name in suffix_removal_av_set:
            label = label.rsplit('.', 1)[0]

        # Truncate after last '.' 
        # if suffix only contains digits or uppercase (no lowercase) chars
        if av_name == 'AVG':
            tokens = label.rsplit('.', 1)
            if len(tokens) > 1 and re.match("^[A-Z0-9]+$", tokens[1]):
                label = tokens[0]

        # Truncate after last '!'
        if av_name == 'Agnitum':
            label = label.rsplit('!', 1)[0]

        return label


    def get_label_tags(self, label, hashes):
        """Return list of tags in given label
 
           Tokenizes label, filters unneeded tokens, and 
           applies tagging rules
        """
        # Initialize set of tags to return
        # We use a set to avoid duplicate tokens in the same AV label
        # This avoids "potentially unwanted" contributing twice BEH:pup
        tags = set()

        # If empty label, nothing to do
        if not label:
            return tags

        # Split label into tokens and process each token
        for token in re.split("[^0-9a-zA-Z]", label):
            # Convert token to lowercase
            token = token.lower()

            # Remove digits at the end
            end_len = len(re.findall("\d*$", token)[0])
            if end_len:
                token = token[:-end_len]

            # Ignore token if prefix of a hash of the sample
            # Most AVs use MD5 prefixes in labels, 
            # but we check SHA1 and SHA256 as well
            hash_token = False
            for hash_str in hashes:
                if hash_str[0:len(token)] == token:
                  hash_token = True
                  break
            if hash_token:
                continue

            # Ignore generic tokens
            if self.taxonomy.is_generic(token):
                continue

            # Apply tagging rule
            dst_l = self.tagging.get_dst(token)
            if dst_l:
                # Ignore generic tokens
                for t in dst_l:
                    if not self.taxonomy.is_generic(t):
                        tags.add(t)
            # Add token if longer than 3 characters and no tagging rule
            elif len(token) > 3:
                tags.add(token)

        # Return tags
        return tags


    def _expand(self, tag_set):
        """Return expanded set of tags"""
        ret = set()
        for t in tag_set:
            # Include tag
            ret.add(t)

            # Include target of expansion rule in output
            ret.update(self.expansions.get_dst(t))

            # Include implicit expansions in taxonomy
            ret.update(self.taxonomy.expand(t))

        # Return a list for backwards compatibility 
        return ret

    def get_sample_tags(self, sample_info, expand=True):
        """Returns dictionary tag -> AV list of tags for the given sample"""

        # Initialize auxiliary data structures
        duplicates = set()
        av_dict = {}

        # Process each AV label
        for (av_name, label) in sample_info.labels:
            # If empty label, nothing to do
            if not label:
                continue

            ################
            # AV selection #
            ################
            if self.avs and av_name not in self.avs:
                continue

            #####################
            # Duplicate removal #
            #####################

            # Emsisoft uses same label as 
            # GData/ESET-NOD32/BitDefender/Ad-Aware/MicroWorld-eScan,
            # but suffixes ' (B)' to their label. Remove the suffix.
            if label.endswith(' (B)'):
                label = label[:-4]

            # F-Secure uses Avira's engine since Nov. 2018
            # but prefixes 'Malware.' to Avira's label. Remove the prefix.
            if label.startswith('Malware.'):
                label = label[8:]

            # Other engines often use exactly the same label, e.g.,
            #   AVG/Avast
            #   K7Antivirus/K7GW
            #   Kaspersky/ZoneAlarm

            # If we have seen the exact same label before, skip
            if label in duplicates:
                continue
            # If not, we add it to duplicates
            else:
                duplicates.add(label)

            ##################
            # Suffix removal #
            ##################
            label = self._remove_suffixes(av_name, label)

            ########################################################
            # Tokenization and tagging                             #
            ########################################################
            hashes = [ sample_info.md5, sample_info.sha1, sample_info.sha256 ]
            tags = self.get_label_tags(label, hashes)

            ########################################################
            # Expansions                                           #
            ########################################################
            if expand:
                expanded_tags = self._expand(tags)
            else:
                expanded_tags = tags

            ########################################################
            # Stores information that relates AV vendors with tags #
            ########################################################
            for t in expanded_tags:
                av_dict.setdefault(t, []).append(av_name)


        return av_dict

    def rank_tags(self, av_dict, threshold=1):
        """Return list of (tag, confidence) ranked by decreasing confidence

           Filters tags with less or equal threshold confidence
        """
        pairs = ((t, len(avs)) for (t,avs) in av_dict.items() 
                    if len(avs) > threshold)
        return sorted(pairs, key=itemgetter(1,0), reverse=True)

    def get_sample_vt_count(self, sample_info):
        """Return number of detections for sample using AV whitelist"""
        if self.avs is None:
            return len(sample_info.labels)
        else:
            cnt = 0
            for (av_name, label) in sample_info.labels:
                if av_name in self.avs:
                    cnt += 1
            return cnt

    def get_family_from_label(self, label):
        """Return family from given label, or None if no family found"""
        fam = None
        tags = self.get_label_tags(label, [])
        for t in tags:
            cat = self.taxonomy.get_category(t)
            if (cat == "UNK") or (cat == "FAM"):
                fam = t
                break
        return fam




#!/usr/bin/env python3

import argparse
import gzip
import json
import logging
import os
import string
import sys
import traceback

from operator import itemgetter

try:
    from avclass import DEFAULT_TAX_PATH, DEFAULT_TAG_PATH, DEFAULT_EXP_PATH
    from avclass.common import AvLabels, Taxonomy, SampleInfo
    from avclass import evaluate as ec
except ModuleNotFoundError:
    # Helps find the avclasses when run from console
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from avclass import DEFAULT_TAX_PATH, DEFAULT_TAG_PATH, DEFAULT_EXP_PATH
    from avclass.common import AvLabels, Taxonomy, SampleInfo
    from avclass import evaluate as ec

# Set logging
log = logging.getLogger(__name__)

# Log warn and above to stderr
formatter = logging.Formatter(u'%(message)s')
handler_stderr = logging.StreamHandler(sys.stderr)
handler_stderr.setLevel(logging.INFO)
handler_stderr.setFormatter(formatter)
root = logging.getLogger()
root.setLevel(logging.DEBUG)
root.addHandler(handler_stderr)

# Default hash to name samples
default_hash_type = "md5"

def guess_hash(h):
    """Guess the hash type of input string"""
    hlen = len(h)
    if hlen == 32:
        return 'md5'
    elif hlen == 40:
        return 'sha1'
    elif hlen == 64:
        return 'sha256'
    else:
        return None

def read_avs(filepath):
    """Read AV engine set from given file"""
    with open(filepath) as fd:
        avs = set(map(str.strip, fd.readlines()))
    return avs

def read_gt(filepath):
    """Read ground truth from given file"""
    gt_dict = {}
    with open(filepath, 'r') as fd:
        for line in fd:
            gt_hash, family = map(str, line.strip().split('\t', 1))
            gt_dict[gt_hash] = family
    return gt_dict

class FileLabeler:
    """Class to extract tags from files"""
    def __init__(self,
        out_fd,
        tag_file = DEFAULT_TAG_PATH,
        exp_file = DEFAULT_EXP_PATH,
        tax_file = DEFAULT_TAX_PATH,
        av_l = None,
        gt_dict = None,
        hash_type = default_hash_type,
        collect_relations = False,
        collect_vendor_info = False,
        collect_stats = False,
        output_all_tags = False,
        output_pup_flag = False,
        output_vt_tags = False
    ):
        """Initialize labeler"""
        # Create AvLabels object
        self.av_labels = AvLabels(tag_file, exp_file, tax_file, av_l=av_l)
        # Store inputs
        self.out_fd = out_fd
        self.gt_dict = gt_dict
        self.hash_type = hash_type
        self.collect_relations = collect_relations
        self.collect_vendor_info = collect_vendor_info
        self.collect_stats = collect_stats
        self.output_all_tags = output_all_tags
        self.output_pup_flag = output_pup_flag
        self.output_vt_tags = output_vt_tags
        # Initialize state
        self.first_token_dict = {}
        self.token_count_map = {}
        self.pair_count_map = {}
        self.vt_all = 0
        self.avtags_dict = {}
        self.stats = {
            'samples': 0,
            'noscans': 0,
            'tagged': 0,
            'maltagged': 0,
            'FAM': 0,
            'CLASS': 0,
            'BEH': 0,
            'FILE': 0,
            'UNK': 0
        }

    @staticmethod
    def get_sample_info_lb(vt_rep):
        """Parse sample information from basic report"""
        return SampleInfo(vt_rep['md5'], vt_rep['sha1'], vt_rep['sha256'],
                          vt_rep['av_labels'], [])

    @staticmethod
    def get_sample_info_vt_v2(vt_rep):
        """Parse sample information from VT v2 report"""
        label_pairs = []
        # Obtain scan results, if available
        try:
            scans = vt_rep['scans']
            md5 = vt_rep['md5']
            sha1 = vt_rep['sha1']
            sha256 = vt_rep['sha256']
        except KeyError:
            return None
        # Obtain labels from scan results
        for av, res in scans.items():
            if res['detected']:
                label = res['result']
                clean_label = ''.join(filter(
                                  lambda x: x in string.printable,
                                    label)).strip()
                label_pairs.append((av, clean_label))
        # Obtain VT tags, if available
        vt_tags = vt_rep.get('tags', [])

        return SampleInfo(md5, sha1, sha256, label_pairs, vt_tags)

    @staticmethod
    def get_sample_info_vt_v3(vt_rep):
        """Parse sample information from VT v3 report"""
        # VT file reports in APIv3 contain all info under 'data'
        # but reports from VT file feed (also APIv3) don't have it
        # Handle both cases silently here
        if 'data' in vt_rep:
            vt_rep = vt_rep['data']
        label_pairs = []
        # Obtain scan results, if available
        try:
            scans = vt_rep['attributes']['last_analysis_results']
            md5 = vt_rep['attributes']['md5']
            sha1 = vt_rep['attributes']['sha1']
            sha256 = vt_rep['attributes']['sha256']
        except KeyError:
            return None
        # Obtain labels from scan results
        for av, res in scans.items():
            label = res['result']
            if label is not None:
                clean_label = ''.join(filter(
                                  lambda x: x in string.printable,
                                    label)).strip()
                label_pairs.append((av, clean_label))
        # Obtain VT tags, if available
        vt_tags = vt_rep['attributes'].get('tags', [])

        return SampleInfo(md5, sha1, sha256, label_pairs, vt_tags)

    @staticmethod
    def get_sample_info_opswat_md(vt_rep):
        """Parse sample information from OPSWAT MetaDefender report"""
        try:
            scans = vt_rep["scan_results"]["scan_details"]
            md5 = vt_rep["file_info"]["md5"]
            sha1 = vt_rep["file_info"]["sha1"]
            sha256 = vt_rep["file_info"]["sha256"]
        except KeyError:
            return None
        # Obtain labels from scan results
        label_pairs = []
        for av, res in scans.items():
            label = res["threat_found"]
            if label is not None and res["scan_result_i"] == 1:
                clean_label = "".join(
                    filter(lambda x: x in string.printable, label)
                ).strip()
                label_pairs.append((av, clean_label))

        return SampleInfo(md5, sha1, sha256, label_pairs, [])

    def open_file(self, filepath):
        """Guess filetype and return file descriptor to file"""
        # Check if file is gzipped by opening it as raw data
        with open(filepath, "rb") as test_fd:
            is_gzipped = test_fd.read(2) == b"\x1f\x8b"
        # Open file correctly
        if is_gzipped:
            fd = gzip.open(filepath, "rt")
        else:
            fd = open(filepath, "r")
        # Read first line
        first_line = fd.readline().strip('\n')
        # Parse line
        report = json.loads(first_line)
        # Check type by parsing the first line
        sample_info = self.get_sample_info_vt_v3(report)
        if sample_info is not None:
            itype = "vt3"
            get_sample_info_fun = self.get_sample_info_vt_v3
        else:
            sample_info = self.get_sample_info_vt_v2(report)
            if sample_info is not None:
                itype = "vt2"
                get_sample_info_fun = self.get_sample_info_vt_v2
            else:
                sample_info = self.get_sample_info_opswat_md(report)
                if sample_info is not None:
                    itype = "md"
                    get_sample_info_fun = self.get_sample_info_opswat_md
                else:
                    itype = "lb"
                    get_sample_info_fun = self.get_sample_info_lb
        # Set file pointer to beginning again
        fd.seek(0, 0)
        # Return file descriptor and type
        return fd, itype, get_sample_info_fun

    def process_line(self, line, get_sample_info):
        """Tag report line and output results"""
        # If blank line, skip
        if line == '\n':
            return

        # Debug info
        if (self.vt_all > 0) and (self.vt_all % 100 == 0):
            sys.stderr.write('\r[-] %d reports read' % self.vt_all)
            sys.stderr.flush()
        self.vt_all += 1

        # Read JSON line
        vt_rep = json.loads(line)

        # Extract sample info
        sample_info = get_sample_info(vt_rep)

        # If no sample info, log error and continue
        if sample_info is None:
            try:
                name = vt_rep['md5']
                log.warning('\nNo scans for %s' % name)
            except KeyError:
                log.warning('\nCould not process: %s' % line)
            self.stats['noscans'] += 1
            return

        # Sample's name is selected hash type (md5 by default)
        name = getattr(sample_info, self.hash_type)

        # If the VT report has no AV labels, output and continue
        if not sample_info.labels:
            self.out_fd.write('%s\t-\t[]\n' % (name))
            return

        # Compute VT_Count (using list of AV engines if provided)
        vt_count = self.av_labels.get_sample_vt_count(sample_info)

        # Get the distinct tokens from all the av labels in the report
        # And print them. 
        try:
            av_tmp = self.av_labels.get_sample_tags(sample_info,
                                          expand=(not self.collect_relations))
            tags = self.av_labels.rank_tags(av_tmp)

            # AV VENDORS PER TOKEN
            if self.collect_vendor_info:
                for t in av_tmp:
                    tmap = self.avtags_dict.get(t, {})
                    for av in av_tmp[t]:
                        ctr = tmap.get(av, 0)
                        tmap[av] = ctr + 1
                    self.avtags_dict[t] = tmap

            if self.collect_relations:
                prev_tokens = set()
                for entry in tags:
                    curr_tok = entry[0]
                    curr_count = self.token_count_map.get(curr_tok, 0)
                    self.token_count_map[curr_tok] = curr_count + 1
                    for prev_tok in prev_tokens:
                        if prev_tok < curr_tok:
                            pair = (prev_tok,curr_tok)
                        else:
                            pair = (curr_tok,prev_tok)
                        pair_count = self.pair_count_map.get(pair, 0)
                        self.pair_count_map[pair] = pair_count + 1
                    prev_tokens.add(curr_tok)

            # Collect stats
            # FIX: should iterate once over tags, 
            # for both stats and collect_relations
            if tags:
                self.stats["tagged"] += 1
                if self.collect_stats:
                    if (vt_count > 3):
                        self.stats["maltagged"] += 1
                        cat_map = {'FAM': False, 'CLASS': False,
                                   'BEH': False, 'FILE': False, 'UNK':
                                       False}
                        for t in tags:
                            path, cat = self.av_labels.taxonomy.get_info(t[0])
                            cat_map[cat] = True
                        for c in cat_map:
                            if cat_map[c]:
                                self.stats[c] += 1

            # Check if sample is PUP, if requested
            if self.output_pup_flag:
                is_pup = self.av_labels.is_pup(tags, self.av_labels.taxonomy)
            else:
                is_pup = None

            # Select family for sample
            fam = "SINGLETON:" + name
            for (t,s) in tags:
                cat = self.av_labels.taxonomy.get_category(t)
                if (cat == "UNK") or (cat == "FAM"):
                    fam = t
                    break

            # Get ground truth family, if available
            if self.gt_dict is not None:
                self.first_token_dict[name] = fam
                gt_family = self.gt_dict.get(name, "")
            else:
                gt_family = None

            # Get VT tags, if requested
            if self.output_vt_tags:
                vt_tags = sample_info.vt_tags
            else:
                vt_tags = None

            # Output sample information
            if self.output_all_tags:
                self.output_with_tags_format(name, vt_count, tags)
            else:
                self.output_with_family_format(name, fam, gt_family, 
                                                is_pup, vt_tags)
        except:
            traceback.print_exc(file=sys.stderr)
            return

    def process_file(self, ifile):
        # Open file
        fd, itype, get_sample_info = self.open_file(ifile)

        # Debug info, file processed
        log.info('[-] Processing input file %s (%s)' % (ifile, itype))

        # Process all lines in file
        for line in fd:
            self.process_line(line, get_sample_info)

        # Debug info
        sys.stderr.write('\r[-] %d reports read' % self.vt_all)
        sys.stderr.flush()
        sys.stderr.write('\n')

        # Close file
        fd.close()

        # Print statistics
        log.info(
                "[-] Samples: %d NoScans: %d NoTags: %d GroundTruth: %d" % (
                    self.vt_all,
                    self.stats['noscans'],
                    self.vt_all - self.stats['tagged'], 
                    len(self.gt_dict) if self.gt_dict else 0)
        )

    def compute_accuracy(self):
        """Compute accuracy statistics using ground truth"""
        return ec.eval_precision_recall_fmeasure(self.gt_dict,
                                                 self.first_token_dict)

    def format_tag_pairs(self, tag_l):
        """Return ranked tags as string"""
        if not tag_l:
            return ""
        if self.av_labels.taxonomy is not None:
            p = self.av_labels.taxonomy.get_path(tag_l[0][0])
        else:
            p = tag_l[0][0]
        out = "%s|%d" % (p, tag_l[0][1])
        for (t,s) in tag_l[1:]:
            if self.av_labels.taxonomy is not None:
                p = self.av_labels.taxonomy.get_path(t)
            else:
                p = t
            out += ",%s|%d" % (p, s)
        return out

    def output_with_family_format(self, name, family, gt_family, 
                                  is_pup, vt_tags):
        """Output sample family results"""
        # Prepare is_pup_str
        if is_pup is not None:
            is_pup_str = "\t1" if is_pup else "\t0"
        else:
            is_pup_str = ""
        # Prepare ground truth family
        if gt_family is not None:
            gt_family = '\t' + gt_family
        else:
            gt_family = ""
        # Prepare vt tags
        vtt = '\t' + ','.join(vt_tags) if vt_tags else ""
        # Write info
        self.out_fd.write('%s\t%s%s%s%s\n' % 
                            (name, family, gt_family, is_pup_str, vtt))

    def output_with_tags_format(self, name, vt_count, tags):
        """Output sample tags"""
        # Prepare tags
        tag_str = self.format_tag_pairs(tags)
        # Write info
        self.out_fd.write('%s\t%d\t%s\n' % (name, vt_count, tag_str))

    def output_relations(self, filepath):
        """Output collected relations to given file"""
        # Open file
        fd = open(filepath, 'w+')
        # Sort token pairs by number of times they appear together
        sorted_pairs = sorted(
            self.pair_count_map.items(), key=itemgetter(1))

        # Output header line
        fd.write("# t1\tt2\t|t1|\t|t2|\t"
                       "|t1^t2|\t|t1^t2|/|t1|\t|t1^t2|/|t2|\n")
        # Compute token pair statistic and output to alias file
        for (t1, t2), c in sorted_pairs:
            n1 = self.token_count_map[t1]
            n2 = self.token_count_map[t2]
            if (n1 < n2):
                x = t1
                y = t2
                xn = n1
                yn = n2
            else:
                x = t2
                y = t1
                xn = n2
                yn = n1
            f = float(c) / float(xn)
            finv = float(c) / float(yn)
            x = self.av_labels.taxonomy.get_path(x)
            y = self.av_labels.taxonomy.get_path(y)
            fd.write("%s\t%s\t%d\t%d\t%d\t%0.2f\t%0.2f\n" % (
                x, y, xn, yn, c, f, finv))
        # Close file
        fd.close()

    def output_stats(self, filepath):
        fd = open(filepath, 'w')
        num_samples = self.vt_all
        fd.write('Samples: %d\n' % num_samples)
        num_tagged = self.stats['tagged']
        frac = float(num_tagged) / float(num_samples) * 100
        fd.write('Tagged (all): %d (%.01f%%)\n' % (num_tagged, frac))
        num_maltagged = self.stats['maltagged']
        frac = float(num_maltagged) / float(num_samples) * 100
        fd.write('Tagged (VT>3): %d (%.01f%%)\n' % (num_maltagged, frac))
        for c in ['FILE','CLASS','BEH','FAM','UNK']:
            count = self.stats[c]
            frac = float(count) / float(num_maltagged) * 100
            fd.write('%s: %d (%.01f%%)\n' % (c, self.stats[c], frac))
        fd.close()

    def output_vendor_info(self, filepath):
        fd = open(filepath, 'w')
        for t in sorted(self.avtags_dict.keys()):
            fd.write('%s\t' % t)
            pairs = sorted(self.avtags_dict[t].items(),
                            key=lambda pair : pair[1],
                            reverse=True)
            for pair in pairs:
                fd.write('%s|%d,' % (pair[0], pair[1]))
            fd.write('\n')
        fd.close()


def main():
    # Parse arguments
    args, ifile_l = parse_args()

    # Read AV engines to be used, if provided
    engine_l = read_avs(args.av) if args.av else None

    # Read ground truth, if provided
    gt_dict = read_gt(args.gt) if args.gt else None

    # Select hash type used to name samples
    if args.gt:
        hash_type = guess_hash(list(gt_dict.keys())[0])
    elif args.hash:
        hash_type = args.hash 
    else:
        hash_type = default_hash_type

    # Set output descriptor
    if args.o:
        out_fd = open(args.o, 'w', encoding='utf-8')
    else:
        out_fd = sys.stdout

    # Create file labeler
    labeler = FileLabeler(
        out_fd,
        tag_file = args.tag,
        exp_file = args.exp,
        tax_file = args.tax,
        av_l = engine_l,
        gt_dict = gt_dict,
        hash_type = hash_type,
        collect_relations = args.aliasdetect,
        collect_vendor_info = args.avtags,
        collect_stats = args.stats,
        output_all_tags = args.t,
        output_pup_flag = args.pup,
        output_vt_tags = args.vtt
    )

    # Process each input file
    for ifile in ifile_l:
        labeler.process_file(ifile)

    # If ground truth, print precision, recall, and F1-measure
    if args.gt:
        precision, recall, fmeasure = labeler.compute_accuracy()
        log.info(
            "Precision: %.2f\tRecall: %.2f\tF1-Measure: %.2f\n" % \
                          (precision, recall, fmeasure))

    # Select output prefix
    out_prefix = os.path.basename(os.path.splitext(ifile_l[0])[0])

    # Output stats
    if args.stats:
        stats_filepath = "%s.stats" % out_prefix
        labeler.output_stats(stats_filepath)
        log.info('[-] Stats in %s' % (stats_filepath))

    # Output vendor info
    if args.avtags:
        vendor_filepath = "%s.avtags" % out_prefix
        labeler.output_vendor_info(vendor_filepath)
        log.info('[-] Vendor info in %s' % (vendor_filepath))

    # If alias detection, print map
    if args.aliasdetect:
        alias_filepath = "%s.alias" % out_prefix
        labeler.output_relations(alias_filepath)
        log.info('[-] Alias data in %s' % (alias_filepath))

    # Close output file
    if args.o:
        out_fd.close()

def parse_args():
    argparser = argparse.ArgumentParser(prog='avclass')

    argparser.add_argument('-f',
        action='append',
        help = 'Input JSONL file with AV labels.')

    argparser.add_argument('-d',
        action='append',
        help = 'Input directory. Process all files in this directory.')

    argparser.add_argument('-t',
        action='store_true',
        help='Output all tags, not only the family.')

    argparser.add_argument('-o',
        help='Output results to this file instead of stdout.')

    argparser.add_argument('-gt',
        help='file with ground truth. '
             'If provided it evaluates clustering accuracy. '
             'Prints precision, recall, F1-measure.')

    argparser.add_argument('-pup',
        action='store_true',
        help='if used each sample is classified as PUP or not')

    argparser.add_argument('-tag',
        default = DEFAULT_TAG_PATH,
        help='file with tagging rules.')

    argparser.add_argument('-tax',
        default = DEFAULT_TAX_PATH,
        help='file with taxonomy.')

    argparser.add_argument('-exp',
        default = DEFAULT_EXP_PATH,
        help='file with expansion rules.')

    argparser.add_argument('-aliasdetect',
        action='store_true',
        help='if used produce aliases file at end')

    argparser.add_argument('-av',
        help='file with list of AVs to use')

    argparser.add_argument('-avtags',
        action='store_true',
        help='extracts tags per av vendor')

    argparser.add_argument('-hash',
        choices=['md5', 'sha1', 'sha256'],
        help='hash used to name samples. Should match ground truth')

    argparser.add_argument('-vtt',
        action='store_true',
        help='Include VT tags in the output.')

    argparser.add_argument('-stats',
        action='store_true',
        help='if used produce 1 file with stats per category '
              '(File, Class, Behavior, Family, Unclassified)')

    args = argparser.parse_args()

    if (not args.f) and (not args.d):
        log.warning('No input files to process. Use -f or -d options')
        sys.exit(1)

    if args.tag:
        if args.tag == '/dev/null':
            log.info('[-] Using no tagging rules')
        else:
            log.info('[-] Using tagging rules in %s' % args.tag)
    else:
        log.info('[-] Using default tagging rules in %s' % DEFAULT_TAG_PATH)

    if args.tax:
        if args.tax == '/dev/null':
            log.info('[-] Using no taxonomy')
        else:
            log.info('[-] Using taxonomy in %s' % args.tax)
    else:
        log.info('[-] Using default taxonomy in %s' % DEFAULT_TAX_PATH)

    if args.exp:
        if args.exp == '/dev/null':
            log.info('[-] Using no expansion tags')
        else:
            log.info('[-] Using expansion tags in %s' % args.exp)
    else:
        log.info('[-] Using default expansion tags in %s' % DEFAULT_EXP_PATH)

    if args.av:
        log.info("[-] Using AV engines in %s" % args.av)

    # Build list of input files
    files = set(args.f) if args.f is not None else {}
    if args.d:
        for d in args.d:
            if os.path.isdir:
                for f in os.listdir(d):
                    filepath = os.path.join(d, f)
                    if os.path.isfile(filepath):
                        files.add(filepath)
            else:
                log.warning('Not a valid directory: %s' % d)
                sys.exit(1)
    ifile_l = sorted(files)

    # Check we have some file to process
    if (not ifile_l):
        log.warning('No input files to process.')
        sys.exit(1)

    return args, ifile_l

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import json
import os
import uuid
import sys

try:
    from avclass import DEFAULT_TAX_PATH, DEFAULT_TAG_PATH, DEFAULT_EXP_PATH
    from avclass.common import Taxonomy, Tagging
except ModuleNotFoundError:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from avclass import DEFAULT_TAX_PATH, DEFAULT_TAG_PATH, DEFAULT_EXP_PATH
    from avclass.common import Taxonomy, Tagging

default_file = "avclass.json"

class Misp:
    """A class to produce the MISP taxonomy"""
    def __init__(self, tax_filepath, tag_filepath):
        self.taxonomy = Taxonomy(tax_filepath)
        self.tagging = Tagging(tag_filepath)

    def uuid_gen(self, tag, prefix='avclass:'):
        """Generate a UUID for given tag"""
        return str(uuid.uuid3(uuid.NAMESPACE_DNS, prefix + tag))

    def output_galaxy_file(self, filepath, version):
        """Output MSIP galaxy to given file"""
        galaxy_uuid = uuid.uuid3(uuid.NAMESPACE_DNS,'avclass_galaxy')
        galaxy_header = {
          "description" : "A malware galaxy from AVClass taxonomy",
          "name" : "AVClass",
          "namespace" : "misp",
          "type" : "avclass",
          "uuid" : str(galaxy_uuid),
          "version" : version
        }
        with open(filepath,'w') as fd:
            fd.write(json.dumps(galaxy_header,indent=2,sort_keys=True))

    def output_cluster_file(self, filepath, version):
        """Output MSIP cluster to given file"""
        entries = []
        dst_map = self.tagging.build_synonymn_map()
        for tag in self.taxonomy:
            # Ignore generics
            if tag.cat == 'GEN':
                continue
            # Get synonyms
            synonyms = dst_map.get(tag.name, {})
            # Build entry
            tag_uuid = self.uuid_gen(tag.name)
            entry_dict = {
                'description' : tag.path,
                'meta' : {
                    'refs' : [],
                    'synonyms' : sorted(synonyms),
                    'type' : []
                },
                'uuid' : tag_uuid,
                'value' : tag.name
            }
            entries.append(entry_dict)
        cluster_uuid = uuid.uuid3(uuid.NAMESPACE_DNS,'avclass_cluster')
        cluster_header = {
            'authors' : ['MaliciaLab'],
            'category' : 'tool',
            'description' : 'AVClass cluster',
            'name' : 'AVClass',
            'source' : 'AVClass',
            'type' : 'avclass',
            'uuid' : str(cluster_uuid),
            'values' : entries,
            'version' : version
        }
        with open(filepath,'w') as fd:
            fd.write(json.dumps(cluster_header,indent=2,sort_keys=True))

def main():
    argparser = argparse.ArgumentParser(
        description='Creates MISP taxonomy.')

    argparser.add_argument('-tag',
        help='tagging file',
        default=DEFAULT_TAG_PATH)

    argparser.add_argument('-tax',
        help='taxonomy file',
        default=DEFAULT_TAX_PATH)

    argparser.add_argument('-o',
        help='output directory',
        default="./")

    argparser.add_argument('-v',
        help='version')

    # Parse arguments
    args = argparser.parse_args()

    # Build MISP object
    misp = Misp(args.tax, args.tag)

    # Create output directories if needed
    misp_dir = "%s/misp/" % args.o
    if (not os.path.exists(misp_dir)):
        os.makedirs(misp_dir)
    galaxy_dir = "%s/galaxy" % misp_dir
    if (not os.path.exists(galaxy_dir)):
        os.makedirs(galaxy_dir)
    cluster_dir = "%s/cluster" % misp_dir
    if (not os.path.exists(cluster_dir)):
        os.makedirs(cluster_dir)

    # Output galaxy file
    galaxy_filepath = "%s/%s" % (galaxy_dir, default_file)
    misp.output_galaxy_file(galaxy_filepath, args.v)

    # Output cluster file
    cluster_filepath = "%s/%s" % (cluster_dir, default_file)
    misp.output_cluster_file(cluster_filepath, args.v)

if __name__ == "__main__":
    main()




import os

AVCLASS_ROOT = os.path.dirname(os.path.abspath(__file__))
DATA_FOLDER = os.path.join(AVCLASS_ROOT, 'data/')

RESOURCE_TAG = "default.tagging"
RESOURCE_TAX = "default.taxonomy"
RESOURCE_EXP = "default.expansion"

DEFAULT_TAX_PATH = os.path.join(DATA_FOLDER, RESOURCE_TAX)
DEFAULT_TAG_PATH = os.path.join(DATA_FOLDER, RESOURCE_TAG)
DEFAULT_EXP_PATH = os.path.join(DATA_FOLDER, RESOURCE_EXP)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import logging
import os
import sys

from collections import namedtuple
from operator import itemgetter
# from Levenshtein import ratio as levenshtein_ratio

try:
    from avclass import DEFAULT_TAX_PATH, DEFAULT_TAG_PATH, DEFAULT_EXP_PATH
    from avclass.common import Taxonomy, Tagging, Expansion
except ModuleNotFoundError:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from avclass import DEFAULT_TAX_PATH, DEFAULT_TAG_PATH, DEFAULT_EXP_PATH
    from avclass.common import Taxonomy, Tagging, Expansion

# Set logging
log = logging.getLogger(__name__)

# Log warn and above to stderr
formatter = logging.Formatter(u'%(message)s')
handler_stderr = logging.StreamHandler(sys.stderr)
handler_stderr.setLevel(logging.INFO)
handler_stderr.setFormatter(formatter)
root = logging.getLogger()
root.setLevel(logging.DEBUG)
root.addHandler(handler_stderr)

# Threshold for string similarity
# sim_threshold = 0.6

# Relation
Rel = namedtuple('Rel', ['t1', 't2', 't1_num', 't2_num', 
                         'nalias_num', 'talias_num', 'tinv_alias_num'])

class Update:
    """Update Module"""
    def __init__(self, rel_filepath, in_taxonomy, in_tagging, in_expansion, 
                    n, t):
        # Initialize inputs
        self._out_taxonomy = in_taxonomy
        self._out_tagging = in_tagging
        self._out_expansion = in_expansion
        self._n = n
        self._t = t
        # Initialize blacklist
        self.blist = in_taxonomy.platform_tags()
        # Maps src -> cnt
        self.src_map = {}
        # Read relations from file
        self.rel_set = self.read_relations(rel_filepath)

    def num_rules(self):
        """Number of relations"""
        return len(self.rel_set)

    def is_weak_rel(self, rel):
        """Whether input relationship is weak""" 
        return ((int(rel.nalias_num) < self._n) or
                (float(rel.talias_num) < self._t))

    def is_blacklisted_rel(self, rel):
        """Whether input relationship is blacklisted"""
        return (rel.t1 in self.blist) or (rel.t2 in self.blist)

    def is_known_rel(self, rel):
        """Whether input relationship is known"""
        t1 = rel.t1
        t2 = rel.t2
        # Known taxonomy relation
        if self._out_taxonomy.overlaps(t1,t2):
            return True
        # Known expansion rule
        t1_dst = self._out_expansion.get_dst(t1)
        t2_dst = self._out_expansion.get_dst(t2)
        if (t2 in t1_dst) or (t1 in t2_dst):
            return True
        # Known tagging rule
        t1_dst = sorted(self._out_tagging.get_dst(t1))
        t2_dst = sorted(self._out_tagging.get_dst(t2))
        if (t2 in t1_dst) or (t1 in t2_dst):
            return True
        # Known alias in tagging
        if t1_dst and (t1_dst == t2_dst):
            return True
        return False

    def add_tag(self, name, path):
        """Add tag to taxonomy if not in tagging"""
        l = self._out_tagging.get_dst(name)
        if (not l):
            self._out_taxonomy.add_tag(path)

    def add_expansion(self, src, dst_l):
        """Add expansion rule fixing destination if src in tagging"""
        # Select source handling aliases
        l = self._out_tagging.get_dst(src)
        if l:
            new_src = l[0]
        else:
            new_src = src
        # Select destinations removing overlaps with existing rule
        l = self._out_expansion.get_dst(src)
        if l:
            l.extend(dst_l)
            target_l = self._out_taxonomy.remove_overlaps(l)
            self._out_expansion.add_rule(new_src, target_l, True)
        else:
            self._out_expansion.add_rule(new_src, dst_l, True)

    def add_alias(self, src, dst, dst_prefix):
        """Add alias relation to taxonomy, tagging"""
        # If src in tagging, use most popular target
        l = self._out_tagging.get_dst(src)
        target = dst
        if l:
            cnt_max = self.src_map[dst]
            for e in l:
                cnt = self.src_map.get(e, 0)
                if cnt > cnt_max:
                    target = e
        # If dst is in tagging, update tagging rule destination, 
        l = self._out_tagging.get_dst(dst)
        if l:
            target_l = l
        # else add dst to taxonomy
        else:
            target_l = [target]
            self._out_taxonomy.add_tag('%s:%s' % (dst_prefix, dst))
        # Remove src from taxonomy
        self._out_taxonomy.remove_tag(src)
        # Replace tagging rule
        self._out_tagging.add_rule(src, target_l, True)

    def is_expansion_rel(self, rel):
        """Whether input relation implies expansion rule"""
        c1 = self._out_taxonomy.get_category(rel.t1)
        c2 = self._out_taxonomy.get_category(rel.t2)
        return (((c1 == "FAM") and (c2 != c1) and (c2 != "UNK")) or
                ((c1 == "CLASS") and ((c2 == "FILE") or (c2 == "BEH"))) or
                ((c1 == "UNK") and ((c2 == "BEH") or (c2 == "CLASS"))))

    def find_expansions(self):
        """Find expansions among relations"""
        acc = []
        for rel in self.rel_set:
            p1 = self._out_taxonomy.get_path(rel.t1)
            p2 = self._out_taxonomy.get_path(rel.t2)
            log.debug("Processing %s\t%s" % (p1, p2))
            # Ignore relations where t1 is an alias
            l = self._out_tagging.get_dst(rel.t1)
            if l:
                log.debug("Ignoring relation for alias %s" % p1)
                continue
            if self.is_expansion_rel(rel):
                self.add_expansion(rel.t1, [rel.t2])
                acc.append(rel)
        for rel in acc:
            self.rel_set.remove(rel)

    #def is_alias_rel(self, rel):
    #    """Whether input relation implies alias rule"""
    #    c1 = self._out_taxonomy.get_category(rel.t1)
    #    c2 = self._out_taxonomy.get_category(rel.t2)
    #    return (((c1 == "UNK") and (c2 == "FAM")) or
    #            ((c1 == "UNK") and (c2 == "UNK")))


    #def find_aliases(self):
    #    """Find aliases among relations"""
    #    for rel in self.rel_set:
    #        c1 = self._out_taxonomy.get_category(rel.t1)
    #        c2 = self._out_taxonomy.get_category(rel.t2)
    #        if self.is_alias_rel(rel):
    #            self.G.add_node(rel.t1)
    #            self.G.add_node(rel.t2)
    #            self.G.add_edge(rel.t1, rel.t2, score=rel.talias_num)
    #    self.output_components("comp")

    def process_relation(self, rel):
        """Process relation and update taxonomy/tagging correspondingly"""
        # Obtain tag info
        t1 = rel.t1
        t2 = rel.t2
        p1,c1 = self._out_taxonomy.get_info(rel.t1)
        p2,c2 = self._out_taxonomy.get_info(rel.t2)

        log.debug("Processing %s\t%s" % (p1, p2))

        # If both directions strong, then equivalent, i.e., alias
        if (float(rel.tinv_alias_num) >= self._t):
            if (c1 != "UNK") and (c2 == "UNK"):
                prefix = p1[0:p1.rfind(':')]
            elif (c1 == "UNK") and (c2 != "UNK"):
                prefix = p2[0:p2.rfind(':')]
            elif (c1 == "UNK") and (c2 == "UNK"):
                prefix = "FAM"
            elif (c1 == c2):
                prefix = p1[0:p1.rfind(':')]
            else:
                log.warning("Equivalent rule with different categories: %s\t%s"
                                % (p1, p2))
                return -1
            self.add_alias(t1, t2, prefix)
            return 1

        # UNK -> FAM : alias-family
        elif (c1 == "UNK") and (c2 == 'FAM'):
            self.add_alias(t1, t2, "FAM")
            return 1

        # UNK -> CLASS : taxonomy-family
        # Return 0 so that expansion handled at end
        elif (c1 == "UNK") and (c2 == 'CLASS'):
            self.add_tag(t1, 'FAM:%s' % t1)
            return 0

        # UNK -> BEH : taxonomy-family
        # Return 0 so that expansion handled at end
        elif (c1 == "UNK") and (c2 == 'BEH'):
            self.add_tag(t1, 'FAM:%s' % t1)
            return 0

        # UNK -> FILE : taxonomy-file
        elif (c1 == "UNK") and (c2 == 'FILE'):
            self.add_tag(t1, '%s:%s' % (p2, t1))
            return 1

        # UNK -> UNK
        elif (c1 == "UNK") and (c2 == "UNK"):
            self.add_alias(t1, t2, "FAM")
            return 1

        # FAM -> UNK : alias-family
        elif (c1 == "FAM") and (c2 == "UNK"):
            self.add_alias(t1, t2, "FAM")
            return 1

         # FILE -> UNK : alias-file
        elif (c1 == "FILE") and (c2 == "UNK"):
            prefix = p1[0:p1.rfind(':')]
            self.add_alias(t1, t2, prefix)
            return 1

        # Same category : alias
        elif (c1 == "FAM") and (c2 == "FAM"):
        #elif c1 == c2:
            prefix = p2[0:p2.rfind(':')]
            self.add_alias(t1, t2, prefix)
            return 1

        # Target unknown
        elif (c2 == "UNK"):
            # If tokens are similar, likely family aliases
            # log.info("Similarity: %.02f" % levenshtein_ratio(t1, t2))
            # if (levenshtein_ratio(t1, t2) > sim_threshold):
            #     prefix = p1[0:p1.rfind(':')]
            #     self.add_alias(t1, t2, prefix)
            #     return 1
            # else:
            #     return 0
            return 0

        # Default: review taxonomy
        else:
            return 0


    def run(self):
        """Identify updates"""
        num_iter = 0
        while self.rel_set:
            # Do a pass in remaining relations
            cnt = 0
            new_set = set()
            log.debug("[-] %03d Processing relations" % num_iter)
            while self.rel_set:
                rel = self.rel_set.pop()
                # If known relation, continue
                if self.is_known_rel(rel):
                    continue

                # Process relation
                result = self.process_relation(rel)

                if result:
                    cnt += 1
                else:
                    new_set.add(rel)

            # Update relation set
            self.rel_set = new_set

            # If no relations processed, finish
            if cnt == 0:
                break
            else:
                num_iter += 1

        # Find aliases
        # self.find_aliases()

        # Find expansions
        log.debug("[-] Finding expansions")
        self.find_expansions()


    def read_relations(self, filepath):
        """Returns relations in file as a set

           Filters weak and blacklisted relations
        """
        rel_set = set()
        with open(filepath, 'r') as fd:
            for line in fd:
                # Ignore comments
                if line.startswith('#'):
                    continue
                # Parse line
                t1, t2, t1_num, t2_num, nalias_num, talias_num, \
                  tinv_alias_num = line.strip().split('\t')
                # Build relation
                rel = Rel(t1, t2, t1_num, t2_num, nalias_num,
                          talias_num, tinv_alias_num)
                # Ignore weak relations
                if self.is_weak_rel(rel):
                    continue
                # Ignore blacklisted relations
                if self.is_blacklisted_rel(rel):
                    continue
                # Ignore known relations
                # NOTE: commented since we check if a 
                # relation is known before processing it
                #if self.is_known_rel(rel):
                #    continue
                # Add relation to set
                rel_set.add(rel)
                # Add to src_map
                self.src_map[rel.t1] = rel.t1_num
                self.src_map[rel.t2] = rel.t2_num

        return rel_set

    def output_relations(self, filepath):
        """Output relations to given file"""
        fd = open(filepath, 'w')
        fd.write("# t1\tt2\t|t1|\t|t2|\t|t1^t2|\t|t1^t2|/|t1|\t"
                  "|t1^t2|/|t2|\n")
        sorted_rules = sorted(self.rel_set, 
                              key=(lambda r: (
                                self._out_taxonomy.get_category(r.t1),
                                self._out_taxonomy.get_category(r.t2))), 
                              reverse=False)
        for rel in sorted_rules:
            p1,c1 = self._out_taxonomy.get_info(rel.t1)
            p2,c2 = self._out_taxonomy.get_info(rel.t2)
            fd.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\n" %(
                p1, p2, rel.t1_num, rel.t2_num, rel.nalias_num, 
                rel.talias_num, rel.tinv_alias_num))
        fd.close()

    def output_rule_stats(self, fd):
        """Output rule statistics to given file descriptor"""
        # Initialize maps for statistics
        self.dst_map = {}
        self.cat_pairs_map = {}
        # Compute rule statistics
        for rel in self.rel_set:
            c1 = self._out_taxonomy.get_category(rel.t1)
            c2 = self._out_taxonomy.get_category(rel.t2)
            self.cat_pairs_map[(c1,c2)] = self.cat_pairs_map.get((c1,
                                                                  c2), 0) + 1
            self.dst_map[rel.t2] = self.dst_map.get(rel.t2, 0) + 1
        # Output statistics
        cat_pairs = sorted(update.cat_pairs_map.items(), key=itemgetter(1,0), 
                            reverse=True)
        for (c1,c2), cnt in cat_pairs:
            fd.write("%s\t%s\t%03d\n" % (c1, c2, cnt))

        # Print dst statistics
        dst_pairs = sorted(update.dst_map.items(), key=itemgetter(1,0), 
                            reverse=False)
        for dst, cnt in dst_pairs:
            fd.write("%s\t%03d\n" % (taxonomy.get_path(dst), cnt))

    def output(self, out_prefix):
        """Output updated taxonomy/tagging/expansions files"""
        if (not out_prefix):
            tax_filepath = DEFAULT_TAX_PATH
            tag_filepath = DEFAULT_TAG_PATH
            exp_filepath = DEFAULT_EXP_PATH
        else:
            tax_filepath = out_prefix + ".taxonomy"
            tag_filepath = out_prefix + ".tagging"
            exp_filepath = out_prefix + ".expansion"
        self._out_taxonomy.to_file(tax_filepath)
        log.info('[-] Output %d taxonomy tags to %s' % (
                        len(self._out_taxonomy), tax_filepath))
        self._out_tagging.expand_all_destinations()
        self._out_tagging.to_file(tag_filepath)
        log.info('[-] Output %d tagging rules to %s' % (
                        len(self._out_tagging), tag_filepath))
        self._out_expansion.to_file(exp_filepath)
        log.info('[-] Output %d expansion rules to %s' % (
                        len(self._out_expansion), exp_filepath))


def main():
    argparser = argparse.ArgumentParser(
        description='''Given a .alias file from the labeler, 
        generates updates for the taxonomy, tagging, and expansion files.''')

    argparser.add_argument('-alias',
        help='input file with alias from labeler. Mandatory.')

    argparser.add_argument('-n',
        help='Minimum number of times that a pair of tokens have been seen.'
             'Default: 20',
        type=int,
        default=20)

    argparser.add_argument('-t',
        help='Minimum percentage of times two tokens appear together.'
             'Default: 0.94',
        type=float,
        default=0.94)

    argparser.add_argument('-o',
        help='output prefix for files')

    argparser.add_argument('-update',
        action='store_true',
        help='update default taxonomy,tagging,expansion files in place')

    argparser.add_argument('-tag',
        help='file with tagging rules.',
        default = DEFAULT_TAG_PATH)

    argparser.add_argument('-tax',
        help='file with taxonomy.',
        default = DEFAULT_TAX_PATH)

    argparser.add_argument('-exp',
        help='file with expansion rules.',
        default = DEFAULT_EXP_PATH)

    argparser.add_argument('-v', '--verbose',
        action='store_true',
        help='verbose, prints debugging statements.')

    # Parse arguments
    args = argparser.parse_args()

    # Check we have the input
    if not args.alias:
        log.error('[-] Please provide an alias file with -alias')
        exit(1)

    # Set logging level
    if (args.verbose):
        handler_stderr.setLevel(logging.DEBUG)

    # Set output prefix
    if args.o:
        out_prefix = args.o
    else:
        out_prefix = os.path.splitext(args.alias)[0]

    # Read taxonomy
    taxonomy = Taxonomy(args.tax)
    log.info('[-] Read %d taxonomy tags from %s' % (
                        len(taxonomy), args.tax))

    # Read tagging rules
    tagging = Tagging(args.tag)
    log.info('[-] Read %d tagging rules from %s' % (
                        len(tagging), args.tag))

    # Read expansion rules
    expansion = Expansion(args.exp)
    log.info('[-] Read %d expansion rules from %s' % (
                        len(expansion), args.exp))

    # Build update object
    update = Update(args.alias, taxonomy, tagging, expansion, args.n, args.t)

    log.info('[-] Processing %d relations satisfying t>=%.2f n>=%d' % (
                        update.num_rules(), args.t, args.n))

    # Output initial rules
    update.output_relations(out_prefix + ".orig.rules")

    # Output initial rules statistics
    # update.output_rule_stats(sys.stderr)

    # Process relations
    update.run()

    # Output updated taxonomy,tagging,expansion
    if args.update:
        update.output(None)
    else:
        update.output(out_prefix)

    # Output final rules
    update.output_relations(out_prefix + ".final.rules")

if __name__ == "__main__":
    main()




#!/usr/bin/env python
import sys

def tp_fp_fn(CORRECT_SET, GUESS_SET):
    """
    INPUT: dictionary with the elements in the cluster from the ground truth
    (CORRECT_SET) and dictionary with the elements from the estimated cluster
    (ESTIMATED_SET).

    OUTPUT: number of True Positives (elements in both clusters), False
    Positives (elements only in the ESTIMATED_SET), False Negatives (elements
    only in the CORRECT_SET).
    """
    tp = 0
    fp = 0
    fn = 0
    for elem in GUESS_SET:
        # True Positives (elements in both clusters)
        if elem in CORRECT_SET:
            tp += 1
        else:
            # False Positives (elements only in the "estimated cluster")
            fp += 1
    for elem in CORRECT_SET:
        if elem not in GUESS_SET:
            # False Negatives (elements only in the "correct cluster")
            fn += 1
    return tp, fp, fn


def eval_precision_recall_fmeasure(GROUNDTRUTH_DICT, ESTIMATED_DICT):
    """
    INPUT: dictionary with the mapping "element:cluster_id" for both the ground
    truth and the ESTIMATED_DICT clustering.

    OUTPUT: average values of Precision, Recall and F-Measure.
    """
    # eval: precision, recall, f-measure
    tmp_precision = 0
    tmp_recall = 0

    # build reverse dictionary of ESTIMATED_DICT
    rev_est_dict = {}
    for k, v in ESTIMATED_DICT.items():
        if v not in rev_est_dict:
            rev_est_dict[v] = { k }
        else:
            rev_est_dict[v].add(k)

    # build reverse dictionary of GROUNDTRUTH_DICT
    gt_rev_dict = {}
    for k, v in GROUNDTRUTH_DICT.items():
        if v not in gt_rev_dict:
            gt_rev_dict[v] = { k }
        else:
            gt_rev_dict[v].add(k)

    
    counter, l = 0, len(ESTIMATED_DICT)

    sys.stderr.write('Calculating precision and recall\n')

    # For each element
    for element in ESTIMATED_DICT:
        
        # Print progress
        if counter % 1000 == 0:
            sys.stderr.write('\r%d out of %d' % (counter, l))
            sys.stderr.flush()
        counter += 1

        # Get elements in the same cluster (for "ESTIMATED_DICT cluster")
        guess_cluster_id = ESTIMATED_DICT[element]

        # Get the list of elements in the same cluster ("correct cluster")
        correct_cluster_id = GROUNDTRUTH_DICT[element]

        # Calculate TP, FP, FN
        tp, fp, fn = tp_fp_fn(gt_rev_dict[correct_cluster_id],
                              rev_est_dict[guess_cluster_id])

        # tmp_precision
        p = 1.0*tp/(tp+fp)
        tmp_precision += p
        # tmp_recall
        r = 1.0*tp/(tp+fn)
        tmp_recall += r
    sys.stderr.write('\r%d out of %d' % (counter, l))
    sys.stderr.write('\n')
    precision = 100.0*tmp_precision/len(ESTIMATED_DICT)
    recall = 100.0*tmp_recall/len(ESTIMATED_DICT)
    fmeasure = (2*precision*recall)/(precision+recall)
    return precision, recall, fmeasure


if __name__ == "__main__":

    # The ground truth.
    # Dictionary with mapping: "element : cluster_id".
    diz_grth = {
        "a": 1,
        "b": 1,
        "c": 2,
        "d": 3
    }

    # An example of an "estimated cluster".
    # Dictionary with mapping: "element : cluster_id".
    diz_estim = {
        "a": 66,
        "b": 'malware',
        "c": 'goodware',
        "d": 'trojan'
    }

    # An example of an "estimated cluster": same partitioning as for the ground
    # truth, but just different cluster labels. Precision == Recall ==
    # F-Measure == 100%.
    # Dictionary with mapping: "element : cluster_id".
    diz_estim_grth = {
        "a": 2,
        "b": 2,
        "c": 66,
        "d": 9
    }

    # a sample where estimated != ground truth
    print("Ground truth")
    print("%8s --> %10s\n" % ("Element", "Cluster_ID"))
    for k, v in diz_grth.items():
        print("%8s --> %10s" % (k, v))
    print("\nEstimated clustering")
    print("%8s --> %10s" % ("Element", "Cluster_ID"))
    for k, v in diz_estim.items():
        print("%8s --> %10s" % (k, v))
    # precision, recall, f-measure
    p, r, f = eval_precision_recall_fmeasure(diz_grth, diz_estim)
    print("\nPrecison: %s%%" % p)
    print("Recall: %s%%" % r)
    print("F-Measure: %s%%" % f)
